# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import six

from testinfra.modules.base import InstanceModule


def int_or_float(value):
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


class _Process(dict):

    def __getattr__(self, key):
        try:
            return self.__getitem__(key)
        except KeyError:
            attrs = self["_module"]._get_process_attribute_by_pid(
                self["pid"], key)
            if attrs["lstart"] != self["lstart"]:
                raise RuntimeError((
                    "Process with pid %s start time changed from %s to %s."
                    " This mean the process you are working on does not not "
                    "exist anymore"
                ) % (self["pid"], self["lstart"], attrs["lstart"]))
            return attrs[key]

    def __repr__(self):
        return "<process %s (pid=%s)>" % (self["comm"], self["pid"])


class Process(InstanceModule):
    """Test Processes attributes

    Processes are selected using ``filter()`` or ``get()``, attributes names
    are described in the `ps(1) man page
    <http://man7.org/linux/man-pages/man1/ps.1.html#STANDARD_FORMAT
    SPECIFIERS>`_.

    >>> master = Process.get(user="root", comm="nginx")
    # Here is the master nginx process (running as root)
    >>> master.args
    'nginx: master process /usr/sbin/nginx -g daemon on; master_process on;'
    # Here are the worker processes (Parent PID = master PID)
    >>> workers = Process.filter(ppid=master.pid)
    >>> len(workers)
    4
    # Nginx don't eat memory
    >>> sum([w.pmem for w in workers])
    0.8
    # But php does !
    >>> sum([p.pmem for p in Process.filter(comm="php5-fpm")])
    19.2

    """

    def filter(self, **filters):
        """Get a list of matching process

        >>> Process.filter(user="root", comm="zsh")
        [<process zsh (pid=2715)>, <process zsh (pid=10502)>, ...]
        """
        match = []
        for attrs in self._get_processes(**filters):
            for key, value in filters.items():
                if six.text_type(attrs[key]) != six.text_type(value):
                    break
            else:
                attrs["_module"] = self
                match.append(_Process(attrs))
        return match

    def get(self, **filters):
        """Get one matching process

        Raise ``RuntimeError`` if no process found or multiple process
        matching filters.
        """
        matches = self.filter(**filters)
        if len(matches) == 0:
            raise RuntimeError("No process found")
        elif len(matches) > 1:
            raise RuntimeError("Multiple process found: %s" % (matches,))
        else:
            return matches[0]

    def _get_processes(self, *values, **filters):
        raise NotImplementedError

    def _get_process_attribute_by_pid(self, pid, name):
        raise NotImplementedError

    @classmethod
    def get_module_class(cls, _backend):
        SystemInfo = _backend.get_module("SystemInfo")
        if SystemInfo.type == "linux" or SystemInfo.type.endswith("bsd"):
            return PosixProcess
        else:
            raise NotImplementedError

    def __repr__(self):
        return "<process>"


class PosixProcess(Process):
    # Should be portable on both Linux and BSD

    def _get_processes(self, **filters):
        cmd = "ps -Aww -o %s"
        attributes = set(["pid", "comm", "pcpu", "pmem"]) | set(filters.keys())

        # Theses attributes contains spaces. Put them at the end of the list
        attributes -= set(["lstart", "args"])
        attributes = sorted(attributes)
        attributes.extend(["lstart", "args"])
        arg = ",".join(attributes)

        procs = []
        # skip first line (header)
        for line in self.check_output(cmd, arg).splitlines()[1:]:
            splitted = line.split()
            attrs = {}
            i = 0
            for i, key in enumerate(attributes[:-2]):
                attrs[key] = int_or_float(splitted[i])
            attrs["lstart"] = " ".join(splitted[i+1:i+6])
            attrs["args"] = " ".join(splitted[i+6:])
            procs.append(attrs)
        return procs

    def _get_process_attribute_by_pid(self, pid, name):
        out = self.check_output(
            "ps -ww -p %s -o lstart,%s", six.text_type(pid), name)
        splitted = out.splitlines()[1].split()
        return {
            "lstart": " ".join(splitted[:5]),
            name: int_or_float(splitted[5]),
        }



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import sys

import pytest

pytestmark = pytest.mark.integration
testinfra_hosts = [
    "%s://debian_wheezy" % (b_type,)
    for b_type in ("ssh", "paramiko", "safe-ssh", "docker")
]

if sys.version_info[0] == 2:
    testinfra_hosts.append("ansible://debian_wheezy")


def test_ssh_package(Package):
    ssh = Package("openssh-server")
    assert ssh.is_installed
    assert ssh.version.startswith("1:6.0p1")


def test_ssh_service(Service):
    ssh = Service("ssh")
    assert ssh.is_running
    assert ssh.is_enabled


def test_systeminfo(SystemInfo):
    assert SystemInfo.type == "linux"
    assert SystemInfo.release == "7"
    assert SystemInfo.distribution == "debian"
    assert SystemInfo.codename is None


def test_user(User):
    user = User("sshd")
    assert user.exists
    assert user.name == "sshd"
    assert user.uid == 101
    assert user.gid == 65534
    assert user.group == "nogroup"
    assert user.gids == [65534]
    assert user.groups == ["nogroup"]
    assert user.shell == "/usr/sbin/nologin"
    assert user.home == "/var/run/sshd"


def test_nonexistent_user(User):
    assert not User("zzzzzzzzzz").exists


def test_current_user(User):
    assert User().name == "root"


def test_group(Group):
    assert Group("root").exists
    assert Group("root").gid == 0


def test_process(Process):
    ssh = Process.get(pid=1)
    assert ssh.ppid == 0
    assert ssh.args == "/usr/sbin/sshd -D"
    assert ssh.comm == "sshd"
    assert ssh.euid == 0



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import datetime

from testinfra.modules.base import Module


class File(Module):
    """Test various files attributes"""

    def __init__(self, path):
        self.path = path
        super(File, self).__init__()

    @property
    def exists(self):
        """Test if file exists

        >>> File("/etc/passwd").exists
        True
        >>> File("/nonexistent").exists
        False

        """
        return self.run_test("test -e %s", self.path).rc == 0

    @property
    def is_file(self):
        return self.run_test("test -f %s", self.path).rc == 0

    @property
    def is_directory(self):
        return self.run_test("test -d %s", self.path).rc == 0

    @property
    def is_pipe(self):
        return self.run_test("test -p %s", self.path).rc == 0

    @property
    def is_socket(self):
        return self.run_test("test -S %s", self.path).rc == 0

    @property
    def is_symlink(self):
        return self.run_test("test -L %s", self.path).rc == 0

    @property
    def linked_to(self):
        """Resolve symlink

        >>> File("/var/lock").linked_to
        '/run/lock'
        """
        return self.check_output("readlink -f %s", self.path)

    @property
    def user(self):
        """Return file owner as string

        >>> File("/etc/passwd").group
        'root'
        """
        raise NotImplementedError

    @property
    def uid(self):
        """Return file user id as integer

        >>> File("/etc/passwd").uid
        0
        """
        raise NotImplementedError

    @property
    def group(self):
        raise NotImplementedError

    @property
    def gid(self):
        raise NotImplementedError

    @property
    def mode(self):
        """Return file mode as octal integer

        >>> File("/etc/passwd").mode
        384  # 0o600 (octal)
        >>> File("/etc/password").mode == 0o600
        True
        >>> oct(File("/etc/password").mode) == '0600'
        True

        Note: Python 3 oct(x)_ function will produce ``'0o600'``

        You can also utilize the file mode constants from
        the stat_ library for testing file mode.

        >>> import stat
        >>> File("/etc/password").mode == stat.S_IRUSR | stat.S_IWUSR
        True

        .. _oct(x): https://docs.python.org/3.5/library/functions.html#oct
        .. _stat: https://docs.python.org/2/library/stat.html
        """
        raise NotImplementedError

    def contains(self, pattern):
        return self.run_test("grep -qs -- %s %s", pattern, self.path).rc == 0

    @property
    def md5sum(self):
        raise NotImplementedError

    @property
    def sha256sum(self):
        raise NotImplementedError

    def _get_content(self, decode):
        out = self.run_test("cat -- %s", self.path)
        if out.rc != 0:
            raise RuntimeError("Unexpected output %s" % (out,))
        if decode:
            return out.stdout
        else:
            return out.stdout_bytes

    @property
    def content(self):
        """Return file content as bytes

        >>> File("/tmp/foo").content
        b'bar'
        """
        return self._get_content(False)

    @property
    def content_string(self):
        """Return file content as string

        >>> File("/tmp/foo").content
        'bar'
        """
        return self._get_content(True)

    @property
    def mtime(self):
        """Return time of last modification as datetime.datetime object

        >>> File("/etc/passwd").mtime
        datetime.datetime(2015, 3, 15, 20, 25, 40)
        """
        raise NotImplementedError

    @property
    def size(self):
        """Return size of file in bytes"""
        raise NotImplementedError

    def __repr__(self):
        return "<file %s>" % (self.path,)

    @classmethod
    def get_module_class(cls, _backend):
        SystemInfo = _backend.get_module("SystemInfo")
        if SystemInfo.type == "linux":
            return GNUFile
        elif SystemInfo.type == "netbsd":
            return NetBSDFile
        elif SystemInfo.type.endswith("bsd"):
            return BSDFile
        else:
            raise NotImplementedError


class GNUFile(File):
    @property
    def user(self):
        return self.check_output("stat -c %%U %s", self.path)

    @property
    def uid(self):
        return int(self.check_output("stat -c %%u %s", self.path))

    @property
    def group(self):
        return self.check_output("stat -c %%G %s", self.path)

    @property
    def gid(self):
        return int(self.check_output("stat -c %%g %s", self.path))

    @property
    def mode(self):
        # Supply a base of 8 when parsing an octal integer
        # e.g. int('644', 8) -> 420
        return int(self.check_output("stat -c %%a %s", self.path), 8)

    @property
    def mtime(self):
        ts = self.check_output("stat -c %%Y %s", self.path)
        return datetime.datetime.fromtimestamp(float(ts))

    @property
    def size(self):
        return int(self.check_output("stat -c %%s %s", self.path))

    @property
    def md5sum(self):
        return self.check_output("md5sum %s | cut -d' ' -f1", self.path)

    @property
    def sha256sum(self):
        return self.check_output(
            "sha256sum %s | cut -d ' ' -f 1", self.path)


class BSDFile(File):
    @property
    def user(self):
        return self.check_output("stat -f %%Su %s", self.path)

    @property
    def uid(self):
        return int(self.check_output("stat -f %%u %s", self.path))

    @property
    def group(self):
        return self.check_output("stat -f %%Sg %s", self.path)

    @property
    def gid(self):
        return int(self.check_output("stat -f %%g %s", self.path))

    @property
    def mode(self):
        # Supply a base of 8 when parsing an octal integer
        # e.g. int('644', 8) -> 420
        return int(self.check_output("stat -f %%Lp %s", self.path), 8)

    @property
    def mtime(self):
        ts = self.check_output("stat -f %%m %s", self.path)
        return datetime.datetime.fromtimestamp(float(ts))

    @property
    def size(self):
        return int(self.check_output("stat -f %%z %s", self.path))

    @property
    def md5sum(self):
        return self.check_output("md5 < %s", self.path)

    @property
    def sha256sum(self):
        return self.check_output(
            "sha256 < %s", self.path)


class NetBSDFile(BSDFile):

    @property
    def sha256sum(self):
        return self.check_output(
            "cksum -a sha256 < %s", self.path)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.modules.base import Module


class Package(Module):
    """Test packages status and version"""

    def __init__(self, name):
        self.name = name
        super(Package, self).__init__()

    @property
    def is_installed(self):
        """Test if the package is installed

        >>> Package("nginx").is_installed
        True

        Supported package systems:

        - apt (Debian, Ubuntu, ...)
        - rpm (RHEL, Centos, Fedora, ...)
        - pkg_info (OpenBSD)
        - pkg_info (NetBSD)
        - pkg (FreeBSD)
        """
        raise NotImplementedError

    @property
    def version(self):
        """Return package version as returned by the package system

        >>> Package("nginx").version
        '1.2.1-2.2+wheezy3'
        """
        raise NotImplementedError

    def __repr__(self):
        return "<package %s>" % (self.name,)

    @classmethod
    def get_module_class(cls, _backend):
        Command = _backend.get_module("Command")
        SystemInfo = _backend.get_module("SystemInfo")
        if SystemInfo.type == "freebsd":
            return FreeBSDPackage
        elif SystemInfo.type in ("openbsd", "netbsd"):
            return OpenBSDPackage
        elif Command.run_test("which apt-get").rc == 0:
            return DebianPackage
        elif Command.run_test("which rpm").rc == 0:
            return RpmPackage
        else:
            raise NotImplementedError


class DebianPackage(Package):

    @property
    def is_installed(self):
        return self.run_test((
            "dpkg-query -f '${Status}' -W %s | "
            "grep -qE '^(install|hold) ok installed$'"), self.name).rc == 0

    @property
    def version(self):
        return self.check_output((
            "dpkg-query -f '${Status} ${Version}' -W %s | "
            "sed -n 's/^install ok installed //p'"), self.name)


class FreeBSDPackage(Package):

    @property
    def is_installed(self):
        EX_UNAVAILABLE = 69
        return self.run_expect(
            [0, EX_UNAVAILABLE], "pkg query %%n %s", self.name).rc == 0

    @property
    def version(self):
        return self.check_output("pkg query %%v %s", self.name)


class OpenBSDPackage(Package):

    @property
    def is_installed(self):
        return self.run_test("pkg_info -e %s", "%s-*" % (self.name,)).rc == 0

    @property
    def version(self):
        out = self.check_output("pkg_info -e %s", "%s-*" % (self.name,))
        # OpenBSD: inst:zsh-5.0.5p0
        # NetBSD: zsh-5.0.7nb1
        return out.split(self.name + "-", 1)[1]


class RpmPackage(Package):

    @property
    def is_installed(self):
        return self.run_test("rpm -q %s", self.name).rc == 0

    @property
    def version(self):
        out = self.check_output("rpm -qi %s", self.name)

        # Name        : bash
        # Version     : 4.2.46
        # ...
        for line in out.splitlines():
            if line.startswith("Version"):
                return line.split(":", 1)[1].strip()
        raise RuntimeError("Cannot parse output '%s'" % (out,))



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals


from testinfra.modules.base import InstanceModule


class Ansible(InstanceModule):
    """Run Ansible module functions

    This module is only available with the :ref:`ansible connection
    backend` connection backend.

    `Check mode
    <https://docs.ansible.com/ansible/playbooks_checkmode.html>`_ is
    enabled by default, you can disable it with `check=False`.

    >>> Ansible("apt", "name=nginx state=present")["changed"]
    False
    >>> Ansible("setup")["ansible_facts"]["ansible_lsb"]["codename"]
    'jessie'
    >>> Ansible("file", "path=/etc/passwd")["mode"]
    '0640'

    """

    def __call__(self, module_name, module_args=None, check=True, **kwargs):
        if not self._backend.HAS_RUN_ANSIBLE:
            raise RuntimeError((
                "Ansible module is only available with ansible "
                "connection backend"))
        return self._backend.run_ansible(
            module_name, module_args, check=check, **kwargs)

    def __repr__(self):
        return "<ansible>"



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import os


def test_file(remote_tempdir, Command, SystemInfo, File):
    filename = os.path.join(remote_tempdir, "f")
    uid, gid = SystemInfo.uid, SystemInfo.gid
    user, group = SystemInfo.user, SystemInfo.group
    assert Command(
        "printf foo > %s && chmod 600 %s && chown %s:%s %s",
        filename, filename, user, group, filename).rc == 0
    f = File(filename)
    assert f.exists
    assert f.is_file
    assert f.content == b"foo"
    assert f.content_string == "foo"
    assert f.user == user
    assert f.uid == uid
    assert f.gid == gid
    assert f.group == group
    assert f.mode == 0o600
    assert f.contains("fo")
    assert not f.is_directory
    assert not f.is_symlink
    assert not f.is_pipe
    assert f.linked_to == filename
    assert f.size == 3
    assert f.md5sum == "acbd18db4cc2f85cedef654fccc4a4d8"
    assert f.sha256sum == (
        "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae"
    )


def test_link(remote_tempdir, Command, File):
    orig = os.path.join(remote_tempdir, "f")
    link = os.path.join(remote_tempdir, "l")
    Command("touch %s", orig)
    assert Command("ln -fsn %s %s", orig, link).rc == 0
    l = File(link)
    assert l.is_symlink
    assert l.is_file
    assert l.linked_to == orig


def test_directory(remote_tempdir, File):
    f = File(remote_tempdir)
    assert f.is_directory
    assert not f.is_file


def test_pipe(remote_tempdir, Command, File):
    filename = os.path.join(remote_tempdir, "f")
    assert Command("mkfifo %s", filename).rc == 0
    assert File(filename).is_pipe


def test_empty_command_output(Command):
    assert Command.check_output("printf ''") == ""


def test_local_command(LocalCommand):
    assert LocalCommand.check_output("true") == ""



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import sys

import pytest

pytestmark = pytest.mark.integration
testinfra_hosts = [
    "%s://fedora_21" % (b_type,)
    for b_type in ("ssh", "paramiko", "safe-ssh", "docker")
]

if sys.version_info[0] == 2:
    testinfra_hosts.append("ansible://fedora_21")


def test_ssh_package(Package):
    ssh = Package("openssh-server")
    assert ssh.is_installed
    assert ssh.version.startswith("6.6")


def test_ssh_service(Service):
    ssh = Service("sshd")
    assert ssh.is_running
    assert ssh.is_enabled


def test_systeminfo(SystemInfo):
    assert SystemInfo.type == "linux"
    assert SystemInfo.release == "21"
    assert SystemInfo.distribution == "fedora"
    assert SystemInfo.codename is None


def test_process(Process):
    systemd = Process.get(pid=1)
    assert systemd.ppid == 0
    assert systemd.args == "/usr/sbin/init"
    assert systemd.comm == "systemd"
    assert systemd.euid == 0



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testinfra.modules.socket import parse_socketspec


def test_parse_socketspec():
    assert parse_socketspec("tcp://22") == ("tcp", None, 22)
    assert parse_socketspec("tcp://:::22") == ("tcp", "::", 22)
    assert parse_socketspec("udp://0.0.0.0:22") == ("udp", "0.0.0.0", 22)
    assert parse_socketspec("unix://can:be.any/thing:22") == (
        "unix", "can:be.any/thing:22", None)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import sys

import pytest

pytestmark = pytest.mark.integration
testinfra_hosts = [
    "%s://debian_jessie" % (b_type,)
    for b_type in ("ssh", "paramiko", "safe-ssh", "docker")
]

if sys.version_info[0] == 2:
    testinfra_hosts.append("ansible://debian_jessie")


def test_ssh_package(Package):
    ssh = Package("openssh-server")
    assert ssh.is_installed
    assert ssh.version.startswith("1:6.7p1")


@pytest.mark.parametrize("name,running,enabled", [
    ("ssh", True, True),
    ("ntp", False, True),
    ("salt-minion", False, False),
])
def test_service(Command, Service, name, running, enabled):

    if name == "ntp":
        # Systemd say no but sysv say yes
        assert Command("systemctl is-enabled ntp").rc == 1

    service = Service(name)
    assert service.is_running == running
    assert service.is_enabled == enabled


def test_systeminfo(SystemInfo):
    assert SystemInfo.type == "linux"
    assert SystemInfo.release[:2] == "8."
    assert SystemInfo.distribution == "debian"
    assert SystemInfo.codename == "jessie"


def test_salt(Salt):
    ssh_version = Salt("pkg.version", "openssh-server", local=True)
    assert ssh_version.startswith("1:6.7p1")


def test_puppet_resource(PuppetResource):
    resource = PuppetResource("package", "openssh-server")
    assert resource["openssh-server"]["ensure"].startswith("1:6.7p1")


def test_facter(Facter):
    assert Facter()["lsbdistcodename"] == "jessie"
    assert Facter("lsbdistcodename") == {
        "lsbdistcodename": "jessie",
    }


def test_sysctl(Sysctl):
    assert Sysctl("kernel.hostname") == "debian_jessie"
    assert isinstance(Sysctl("kernel.panic"), int)


def test_encoding(testinfra_backend, Command):
    if testinfra_backend.get_connection_type() == "ansible":
        pytest.skip("ansible handle encoding himself")

    # jessie image is fr_FR@ISO-8859-15
    cmd = Command("ls -l %s", "/é")
    assert cmd.command == b"ls -l '/\xe9'"
    if testinfra_backend.get_connection_type() == "docker":
        # docker bug ?
        assert cmd.stderr_bytes == (
            b"ls: impossible d'acc\xe9der \xe0 /\xef\xbf\xbd: "
            b"Aucun fichier ou dossier de ce type\n"
        )
    else:
        assert cmd.stderr_bytes == (
            b"ls: impossible d'acc\xe9der \xe0 /\xe9: "
            b"Aucun fichier ou dossier de ce type\n"
        )
        assert cmd.stderr == (
            "ls: impossible d'accéder à /é: "
            "Aucun fichier ou dossier de ce type\n"
        )


def test_socket(testinfra_backend, Socket):
    listening = Socket.get_listening_sockets()
    for spec in (
        "tcp://0.0.0.0:22",
        "tcp://:::22",
        "unix:///run/systemd/private",
    ):
        assert spec in listening
    for spec in (
        "tcp://22",
        "tcp://0.0.0.0:22",
        "tcp://127.0.0.1:22",
        "tcp://:::22",
        "tcp://::1:22",
    ):
        socket = Socket(spec)
        assert socket.is_listening

    if not testinfra_backend.get_connection_type() == "docker":
        for spec in (
            "tcp://22",
            "tcp://0.0.0.0:22",
        ):
            assert len(Socket(spec).clients) >= 1
    assert not Socket("tcp://4242").is_listening


def test_ansible_module(testinfra_backend, Ansible):
    if not testinfra_backend.get_connection_type() == "ansible":
        with pytest.raises(RuntimeError) as excinfo:
            setup = Ansible("setup")
        assert (
            'Ansible module is only available with ansible '
            'connection backend') in str(excinfo.value)
    else:
        setup = Ansible("setup")["ansible_facts"]
        assert setup["ansible_lsb"]["codename"] == "jessie"
        passwd = Ansible("file", "path=/etc/passwd")
        assert passwd["changed"] is False
        assert passwd["gid"] == 0
        assert passwd["group"] == "root"
        assert passwd["mode"] == "0644"
        assert passwd["owner"] == "root"
        assert passwd["size"] == 1369
        assert passwd["path"] == "/etc/passwd"
        assert passwd["state"] == "file"
        assert passwd["uid"] == 0


def test_process(Process):
    systemd = Process.get(pid=1)
    assert systemd.ppid == 0
    assert systemd.args == "/sbin/init"
    assert systemd.comm == "systemd"
    assert systemd.euid == 0



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import logging

import pytest
import testinfra
from testinfra import modules

File = modules.File.as_fixture()
Command = modules.Command.as_fixture()
Package = modules.Package.as_fixture()
Group = modules.Group.as_fixture()
Interface = modules.Interface.as_fixture()
Command = modules.Command.as_fixture()
Service = modules.Service.as_fixture()
SystemInfo = modules.SystemInfo.as_fixture()
User = modules.User.as_fixture()
Salt = modules.Salt.as_fixture()
PuppetResource = modules.PuppetResource.as_fixture()
Facter = modules.Facter.as_fixture()
Sysctl = modules.Sysctl.as_fixture()
Socket = modules.Socket.as_fixture()
Ansible = modules.Ansible.as_fixture()
Process = modules.Process.as_fixture()


@pytest.fixture(scope="module")
def LocalCommand(testinfra_backend):
    """Run commands locally

    Same as `Command` but run commands locally with subprocess even
    when the connection backend is not "local".

    Note: `LocalCommand` does NOT respect ``--sudo`` option
    """
    return testinfra.get_backend("local://").get_module("Command")


@pytest.fixture(scope="module")
def TestinfraBackend(testinfra_backend):
    return testinfra_backend


def pytest_addoption(parser):
    group = parser.getgroup("testinfra")
    group._addoption(
        "--connection",
        action="store",
        dest="connection",
        help=(
            "Remote connection backend (paramiko, ssh, safe-ssh, "
            "salt, docker, ansible)"
        )
    )
    group._addoption(
        "--hosts",
        action="store",
        dest="hosts",
        help="Hosts list (comma separated)",
    )
    group._addoption(
        "--ssh-config",
        action="store",
        dest="ssh_config",
        help="SSH config file",
    )
    group._addoption(
        "--sudo",
        action="store_true",
        dest="sudo",
        help="Use sudo",
    )
    group._addoption(
        "--ansible-inventory",
        action="store",
        dest="ansible_inventory",
        help="Ansible inventory file",
    )
    group._addoption(
        "--nagios",
        action="store_true",
        dest="nagios",
        help="Nagios plugin",
    )


def pytest_generate_tests(metafunc):
    if "testinfra_backend" in metafunc.fixturenames:
        if metafunc.config.option.hosts is not None:
            hosts = metafunc.config.option.hosts.split(",")
        elif hasattr(metafunc.module, "testinfra_hosts"):
            hosts = metafunc.module.testinfra_hosts
        else:
            hosts = [None]
        params = testinfra.get_backends(
            hosts,
            connection=metafunc.config.option.connection,
            ssh_config=metafunc.config.option.ssh_config,
            sudo=metafunc.config.option.sudo,
            ansible_inventory=metafunc.config.option.ansible_inventory,
        )
        ids = [e.get_pytest_id() for e in params]
        metafunc.parametrize(
            "testinfra_backend", params, ids=ids, scope="module")


def pytest_configure(config):
    if config.option.verbose > 0:
        logging.basicConfig()
        logging.getLogger("testinfra").setLevel(logging.DEBUG)



# -*- coding: utf8 -*-
# Copyright © 2015-2016 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from __future__ import absolute_import

import logging

from testinfra.backend import base
import testinfra.utils.ansible_runner as ansible_runner

logger = logging.getLogger("testinfra")


class AnsibleBackend(base.BaseBackend):
    NAME = "ansible"
    HAS_RUN_ANSIBLE = True

    def __init__(self, host, ansible_inventory=None, *args, **kwargs):
        self.host = host
        self.ansible_inventory = ansible_inventory
        super(AnsibleBackend, self).__init__(host, *args, **kwargs)

    def run(self, command, *args):
        command = self.get_command(command, *args)
        out = self.run_ansible("shell", module_args=command)

        # Ansible return an unicode object but this is bytes ...
        # A simple test case is:
        # >>> assert File("/bin/true").content == open("/bin/true").read()
        stdout_bytes = b"".join((chr(ord(c)) for c in out['stdout']))
        stderr_bytes = b"".join((chr(ord(c)) for c in out['stderr']))

        result = base.CommandResult(
            self, out['rc'],
            stdout_bytes,
            stderr_bytes,
            command,
            stdout=out["stdout"], stderr=out["stderr"],
        )
        logger.info("RUN %s", result)
        return result

    def run_ansible(self, module_name, module_args=None, **kwargs):
        return ansible_runner.run(
            self.host, module_name, module_args,
            host_list=self.ansible_inventory,
            **kwargs)

    @classmethod
    def get_hosts(cls, host, **kwargs):
        return ansible_runner.get_hosts(kwargs.get("ansible_inventory"), host)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from __future__ import absolute_import

import logging
import os

try:
    import paramiko
except ImportError:
    HAS_PARAMIKO = False
else:
    HAS_PARAMIKO = True

    class IgnorePolicy(paramiko.MissingHostKeyPolicy):
        """Policy for ignoring missing host key."""
        def missing_host_key(self, client, hostname, key):
            pass

from testinfra.backend import base
logger = logging.getLogger("testinfra")


class ParamikoBackend(base.BaseBackend):
    NAME = "paramiko"

    def __init__(self, hostspec, ssh_config=None, *args, **kwargs):
        self.host, self.user, self.port = self.parse_hostspec(hostspec)
        self.ssh_config = ssh_config
        self._client = None
        super(ParamikoBackend, self).__init__(self.host, *args, **kwargs)

    @property
    def client(self):
        if self._client is None:
            if not HAS_PARAMIKO:
                raise RuntimeError((
                    "You must install paramiko package (pip install paramiko) "
                    "to use the paramiko backend"))
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.WarningPolicy())
            cfg = {
                "hostname": self.host,
                "port": int(self.port) if self.port else 22,
                "username": self.user,
            }
            if self.ssh_config:
                ssh_config = paramiko.SSHConfig()
                with open(self.ssh_config) as f:
                    ssh_config.parse(f)

                for key, value in ssh_config.lookup(self.host).items():
                    if key == "hostname":
                        cfg[key] = value
                    elif key == "user":
                        cfg["username"] = value
                    elif key == "port":
                        cfg[key] = int(value)
                    elif key == "identityfile":
                        cfg["key_filename"] = os.path.expanduser(value[0])
                    elif key == "stricthostkeychecking" and value == "no":
                        client.set_missing_host_key_policy(IgnorePolicy())

            client.connect(**cfg)
            self._client = client
        return self._client

    def run(self, command, *args, **kwargs):
        command = self.get_command(command, *args)
        chan = self.client.get_transport().open_session()
        command = self.encode(command)
        chan.exec_command(command)
        rc = chan.recv_exit_status()
        stdout = b''.join(chan.makefile('rb'))
        stderr = b''.join(chan.makefile_stderr('rb'))
        result = base.CommandResult(self, rc, stdout, stderr, command)
        logger.info("RUN %s", result)
        return result



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.backend import base


class LocalBackend(base.BaseBackend):
    NAME = "local"

    def __init__(self, *args, **kwargs):
        super(LocalBackend, self).__init__("local", **kwargs)

    def get_pytest_id(self):
        return "local"

    @classmethod
    def get_hosts(cls, host, **kwargs):
        return [host]

    def run(self, command, *args, **kwargs):
        if self.sudo:
            command = "sudo " + command
        return self.run_local(command, *args)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.modules.base import Module


class Group(Module):
    """Test unix group"""

    def __init__(self, name=None):
        self.name = name
        super(Group, self).__init__()

    @property
    def exists(self):
        return self.run_expect([0, 2], "getent group %s", self.name).rc == 0

    @property
    def gid(self):
        return int(self.check_output(
            "getent group %s | cut -d':' -f3", self.name))

    def __repr__(self):
        return "<group %s>" % (self.name,)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import locale
import logging
import pipes
import subprocess

import testinfra.modules

logger = logging.getLogger("testinfra")


class CommandResult(object):

    def __init__(
        self, backend, exit_status, stdout_bytes, stderr_bytes, command,
        stdout=None, stderr=None,
    ):
        self.exit_status = exit_status
        self.stdout_bytes = stdout_bytes
        self.stderr_bytes = stderr_bytes
        self._stdout = stdout
        self._stderr = stderr
        self.command = command
        self._backend = backend
        super(CommandResult, self).__init__()

    @property
    def rc(self):
        return self.exit_status

    @property
    def stdout(self):
        if self._stdout is None:
            self._stdout = self._backend.decode(self.stdout_bytes)
        return self._stdout

    @property
    def stderr(self):
        if self._stderr is None:
            self._stderr = self._backend.decode(self.stderr_bytes)
        return self._stderr

    def __repr__(self):
        return (
            "CommandResult(command=%s, exit_status=%s, stdout=%s, "
            "stderr=%s)"
        ) % (
            repr(self.command),
            self.exit_status,
            repr(self.stdout_bytes),
            repr(self.stderr_bytes),
        )


class BaseBackend(object):
    """Represent the connection to the remote or local system"""
    NAME = None
    HAS_RUN_SALT = False
    HAS_RUN_ANSIBLE = False

    def __init__(self, hostname, sudo=False, *args, **kwargs):
        self._encoding = None
        self._module_cache = {}
        self.hostname = hostname
        self.sudo = sudo
        super(BaseBackend, self).__init__()

    @classmethod
    def get_connection_type(cls):
        """Return the connection backend used as string.

        Can be local, paramiko, ssh, docker, salt or ansible
        """
        return cls.NAME

    def get_hostname(self):
        """Return the hostname (for testinfra) of the remote or local system


        Can be useful for multi-hosts tests:

        Example:
        ::

            import requests


            def test(TestinfraBackend):
                host = TestinfraBackend.get_hostname()
                response = requests.get("http://" + host)
                assert response.status_code == 200


        ::

            $ testinfra --hosts=server1,server2 test.py

            test.py::test[paramiko://server1] PASSED
            test.py::test[paramiko://server2] PASSED
        """
        return self.hostname

    def get_pytest_id(self):
        return self.get_connection_type() + "://" + self.get_hostname()

    @classmethod
    def get_hosts(cls, host, **kwargs):
        if host is None:
            raise RuntimeError(
                "One or more hosts is required with the %s backend" % (
                    cls.get_connection_type(),),
            )
        return [host]

    @staticmethod
    def quote(command, *args):
        if args:
            return command % tuple(pipes.quote(a) for a in args)
        else:
            return command

    def get_command(self, command, *args):
        if self.sudo:
            command = self.quote("sudo /bin/sh -c %s", command)
        return self.quote(command, *args)

    def run(self, command, *args, **kwargs):
        raise NotImplementedError

    def run_local(self, command, *args):
        command = self.quote(command, *args)
        command = self.encode(command)
        p = subprocess.Popen(
            command, shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        result = CommandResult(
            self, p.returncode, stdout, stderr, command,
        )
        logger.info("RUN %s", result)
        return result

    @staticmethod
    def parse_hostspec(hostspec):
        host = hostspec
        user = None
        port = None
        if "@" in host:
            user, host = host.split("@", 1)
        if ":" in host:
            host, port = host.split(":", 1)
        return host, user, port

    def get_encoding(self):
        cmd = self.run(
            "python -c 'import locale;print(locale.getpreferredencoding())'")
        if cmd.rc == 0:
            encoding = cmd.stdout_bytes.splitlines()[0].decode("ascii")
        else:
            # Python is not installed, we hope the encoding to be the same as
            # local machine...
            encoding = locale.getpreferredencoding()
        return encoding

    @property
    def encoding(self):
        if self._encoding is None:
            self._encoding = self.get_encoding()
        return self._encoding

    def decode(self, data):
        try:
            return data.decode("ascii")
        except UnicodeDecodeError:
            return data.decode(self.encoding)

    def encode(self, data):
        try:
            return data.encode("ascii")
        except UnicodeEncodeError:
            return data.encode(self.encoding)

    def get_module(self, name):
        """Return the testinfra module adapted to the current backend

        ::

            def test(Package):
                [...]

            # Is equivalent to
            def test(TestinfraBackend):
                Package = TestinfraBackend.get_module("Package")

        """
        try:
            module = self._module_cache[name]
        except KeyError:
            module = getattr(testinfra.modules, name).get_module(self)
            self._module_cache[name] = module
        return module



# -*- coding: utf8 -*-
# Copyright © 2015-2016 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=import-error

from __future__ import print_function

import fileinput
import sys

import ansible
import paramiko


def ssh_config_to_ansible_inventory():
    if len(sys.argv) >= 2:
        stream = fileinput.input([sys.argv[1]])
    else:
        stream = fileinput.input()

    ssh_config = paramiko.SSHConfig()
    ssh_config.parse(stream)

    ansible_major_version = int(ansible.__version__.split(".", 1)[0])
    if ansible_major_version == 1:
        key_map = {
            "hostname": "ansible_ssh_host",
            "user": "ansible_ssh_user",
            "port": "ansible_ssh_port",
            "identityfile": "ansible_ssh_private_key_file",
        }
    elif ansible_major_version == 2:
        key_map = {
            "hostname": "ansible_host",
            "user": "ansible_user",
            "port": "ansible_port",
            "identityfile": "ansible_ssh_private_key_file",
        }
    else:
        raise RuntimeError(
            "Ansible version {} not supported".format(ansible.__version__))

    for hostname in (e for e in ssh_config.get_hostnames() if e != "*"):
        items = [hostname]
        for key, value in ssh_config.lookup(hostname).items():
            if key in key_map:
                if key == "identityfile":
                    value = value[0]
                items.append("{}={}".format(key_map[key], value))
        print(" ".join(items))

if __name__ == "__main__":
    ssh_config_to_ansible_inventory()



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import pytest


@pytest.fixture(scope="function")
def remote_tempdir(request, Command):
    dirname = "/tmp/testinfra-%s" % uuid.uuid4().hex
    assert Command("rm -rf %s; mkdir -p %s", dirname, dirname).rc == 0

    def fin():
        Command("rm -rf %s", dirname)

    request.addfinalizer(fin)
    return dirname


def pytest_addoption(parser):
    parser.addoption("--integration", action="store_true",
                     help="Run integration tests")


def pytest_runtest_setup(item):
    if (
        "integration" in item.keywords and
        not item.config.getoption("--integration")
    ):
        pytest.skip("integration test not selected")



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.modules.base import InstanceModule


class Command(InstanceModule):
    """Run given command and return rc (exit status), stdout and stderr

    >>> cmd = Command("ls -l /etc/passwd")
    >>> cmd.rc
    0
    >>> cmd.stdout
    '-rw-r--r-- 1 root root 1790 Feb 11 00:28 /etc/passwd\\n'
    >>> cmd.stderr
    ''


    Good practice: always use shell arguments quoting to avoid shell injection


    >>> cmd = Command("ls -l -- %s", "/;echo inject")
    CommandResult(
        rc=2, stdout='',
        stderr='ls: cannot access /;echo inject: No such file or directory\\n',
        command="ls -l '/;echo inject'")
    """

    def __call__(self, command, *args, **kwargs):
        return self.run(command, *args, **kwargs)

    def __repr__(self):
        return "<command>"



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from __future__ import absolute_import

import logging

try:
    import salt.client
except ImportError:
    HAS_SALT = False
else:
    HAS_SALT = True

from testinfra.backend import base
logger = logging.getLogger("testinfra")


class SaltBackend(base.BaseBackend):
    HAS_RUN_SALT = True
    NAME = "salt"

    def __init__(self, host, *args, **kwargs):
        self.host = host
        self._client = None
        super(SaltBackend, self).__init__(self.host, *args, **kwargs)

    @staticmethod
    def _check_salt():
        if not HAS_SALT:
            raise RuntimeError(
                "You must install salt package to use the salt backend")

    @property
    def client(self):
        if self._client is None:
            self._check_salt()
            self._client = salt.client.LocalClient()
        return self._client

    def run(self, command, *args):
        command = self.get_command(command, *args)
        out = self.run_salt("cmd.run_all", [command])
        result = base.CommandResult(
            self, out['retcode'], out['stdout'], out['stderr'], command)
        logger.info("RUN %s", result)
        return result

    def run_salt(self, func, args=None):
        out = self.client.cmd(self.host, func, args or [])
        if self.host not in out:
            raise RuntimeError(
                "Error while running %s(%s): %s. Minion not connected ?" % (
                    func, args, out))
        return out[self.host]

    @classmethod
    def get_hosts(cls, host, **kwargs):
        if host is None:
            host = "*"
        if any([c in host for c in "@*[?"]):
            cls._check_salt()
            client = salt.client.LocalClient()
            if "@" in host:
                hosts = client.cmd(
                    host, "test.true", expr_form="compound").keys()
            else:
                hosts = client.cmd(host, "test.true").keys()
            if not hosts:
                raise RuntimeError("No host matching '%s'" % (host,))
            else:
                return hosts
        else:
            return super(SaltBackend, cls).get_hosts(host, **kwargs)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.modules.base import Module


class User(Module):
    """Test unix users

    If name is not supplied, test the current user
    """

    def __init__(self, name=None):
        self._name = name
        super(User, self).__init__()

    @property
    def name(self):
        """Return user name"""
        if self._name is None:
            self._name = self.check_output("id -nu")
        return self._name

    @property
    def exists(self):
        return self.run_test("id %s", self.name).rc == 0

    @property
    def uid(self):
        """Return user ID"""
        return int(self.check_output("id -u %s", self.name))

    @property
    def gid(self):
        """Return effective group ID"""
        return int(self.check_output("id -g %s", self.name))

    @property
    def group(self):
        """Return effective group name"""
        return self.check_output("id -ng %s", self.name)

    @property
    def gids(self):
        """Return the list of user group IDs"""
        return [int(gid) for gid in self.check_output(
            "id -G %s", self.name,
        ).split(" ")]

    @property
    def groups(self):
        """Return the list of user group names"""
        return self.check_output("id -nG %s", self.name).split(" ")

    @property
    def home(self):
        """Return the user home directory"""
        return self.check_output("getent passwd %s", self.name).split(":")[5]

    @property
    def shell(self):
        """Return the user login shell"""
        return self.check_output("getent passwd %s", self.name).split(":")[6]

    def __repr__(self):
        return "<user %s>" % (self.name,)



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from six.moves import urllib

from testinfra.backend import ansible
from testinfra.backend import docker
from testinfra.backend import local
from testinfra.backend import paramiko
from testinfra.backend import salt
from testinfra.backend import ssh

BACKENDS = dict((klass.get_connection_type(), klass) for klass in (
    local.LocalBackend,
    ssh.SshBackend,
    ssh.SafeSshBackend,
    paramiko.ParamikoBackend,
    salt.SaltBackend,
    docker.DockerBackend,
    ansible.AnsibleBackend,
))


def get_backend_class(connection):
    try:
        return BACKENDS[connection]
    except KeyError:
        raise RuntimeError("Unknown connection type '%s'" % (connection,))


def parse_hostspec(hostspec):
    kw = {}
    if hostspec is not None and "://" in hostspec:
        url = urllib.parse.urlparse(hostspec)
        kw["connection"] = url.scheme
        host = url.netloc
        query = urllib.parse.parse_qs(url.query)
        if query.get("sudo", ["false"])[0].lower() == "true":
            kw["sudo"] = True
        for key in (
            "ssh_config", "ansible_inventory",
        ):
            if key in query:
                kw[key] = query.get(key)[0]
    else:
        host = hostspec
    return host, kw


def get_backend(hostspec, **kwargs):
    host, kw = parse_hostspec(hostspec)
    for k, v in kwargs.items():
        kw.setdefault(k, v)
    kw.setdefault("connection", "paramiko")
    klass = get_backend_class(kw["connection"])
    if kw["connection"] == "local":
        return klass(**kw)
    else:
        return klass(host, **kw)


def get_backends(hosts, **kwargs):
    backends = []
    for hostspec in hosts:
        host, kw = parse_hostspec(hostspec)
        for k, v in kwargs.items():
            kw.setdefault(k, v)
        connection = kw.get("connection")
        if host is None and connection is None:
            connection = "local"
        elif host is not None and connection is None:
            connection = "paramiko"
        klass = get_backend_class(connection)
        for name in klass.get_hosts(host, **kw):
            if connection == "local":
                backends.append(klass(**kw))
            else:
                backends.append(klass(name, **kw))
    return backends



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.modules.base import InstanceModule


class SystemInfo(InstanceModule):
    """Return system informations"""

    def __init__(self):
        self._sysinfo = None
        super(SystemInfo, self).__init__()

    @property
    def sysinfo(self):
        if self._sysinfo is None:
            self._sysinfo = self.get_system_info()
        return self._sysinfo

    def get_system_info(self):
        sysinfo = {
            "type": None,
            "distribution": None,
            "codename": None,
            "release": None,
        }
        sysinfo["type"] = self.check_output("uname -s").lower()
        if sysinfo["type"] == "linux":
            lsb = self.run("lsb_release -a")
            if lsb.rc == 0:
                for line in lsb.stdout.splitlines():
                    key, value = line.split(":", 1)
                    key = key.strip().lower()
                    value = value.strip().lower()
                    if key == "distributor id":
                        sysinfo["distribution"] = value
                    elif key == "release":
                        sysinfo["release"] = value
                    elif key == "codename":
                        sysinfo["codename"] = value
            else:
                os_release = self.run("cat /etc/os-release")
                if os_release.rc == 0:
                    for line in os_release.stdout.splitlines():
                        for key, attname in (
                            ("ID=", "distribution"),
                            ("VERSION_ID=", "release"),
                        ):
                            if line.startswith(key):
                                sysinfo[attname] = (
                                    line[len(key):].replace('"', "").
                                    replace("'", "").strip())
        else:
            sysinfo["release"] = self.check_output("uname -r")
            sysinfo["distribution"] = sysinfo["type"]
            sysinfo["codename"] = None
        return sysinfo

    @property
    def type(self):
        """OS type

        >>> SystemInfo.type
        'linux'
        """
        return self.sysinfo["type"]

    @property
    def distribution(self):
        """Distribution name

        >>> SystemInfo.distribution
        'debian'
        """
        return self.sysinfo["distribution"]

    @property
    def release(self):
        """Distribution release number

        >>> SystemInfo.release
        '7.8'
        """
        return self.sysinfo["release"]

    @property
    def codename(self):
        """Release code name

        >>> SystemInfo.codename
        'wheezy'
        """
        return self.sysinfo["codename"]

    @property
    def user(self):
        return self.check_output("id -nu")

    @property
    def uid(self):
        return int(self.check_output("id -u"))

    @property
    def group(self):
        return self.check_output("id -ng")

    @property
    def gid(self):
        return int(self.check_output("id -g"))

    @property
    def hostname(self):
        return self.check_output("hostname -s")



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra import backend

__all__ = ["get_backend", "get_backends"]


_BACKEND_CACHE = {}
_BACKENDS_CACHE = {}


def get_backend(hostspec, **kwargs):
    key = (hostspec, frozenset(kwargs.items()))
    if key not in _BACKEND_CACHE:
        _BACKEND_CACHE[key] = backend.get_backend(hostspec, **kwargs)
    return _BACKEND_CACHE[key]


def get_backends(hosts, **kwargs):
    key = (frozenset(hosts), frozenset(kwargs.items()))
    if key not in _BACKENDS_CACHE:
        _BACKENDS_CACHE[key] = backend.get_backends(hosts, **kwargs)
    return _BACKENDS_CACHE[key]



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

import pytest


class Module(object):
    _backend = None

    def run(self, command, *args, **kwargs):
        return self._backend.run(command, *args, **kwargs)

    def run_expect(self, expected, command, *args, **kwargs):
        """Run command and check it return an expected exit status

        :param expected: A list of expected exit status
        :raises: AssertionError
        """
        __tracebackhide__ = True  # pylint: disable=unused-variable
        out = self.run(command, *args, **kwargs)
        if out.rc not in expected:
            pytest.fail("Unexpected exit code %s for %s" % (out.rc, out))
        return out

    def run_test(self, command, *args, **kwargs):
        """Run command and check it return an exit status of 0 or 1

        :raises: AssertionError
        """
        return self.run_expect([0, 1], command, *args, **kwargs)

    def check_output(self, command, *args, **kwargs):
        """Get stdout of a command which has run successfully

        :returns: stdout without trailing newline
        :raises: AssertionError
        """
        __tracebackhide__ = True  # pylint: disable=unused-variable
        out = self.run(command, *args, **kwargs)
        if out.rc != 0:
            pytest.fail("Unexpected exit code %s for %s" % (out.rc, out))

        if len(out.stdout) and out.stdout[-1] == "\n":
            return out.stdout[:-1]
        else:
            return out.stdout

    @classmethod
    def get_module(cls, _backend):
        klass = cls.get_module_class(_backend)
        return type(klass.__name__, (klass,), {
            "_backend": _backend,
        })

    @classmethod
    def get_module_class(cls, _backend):
        return cls

    @classmethod
    def as_fixture(cls):
        @pytest.fixture(scope="module")
        def f(testinfra_backend):
            return testinfra_backend.get_module(cls.__name__)
        f.__doc__ = cls.__doc__
        return f


class InstanceModule(Module):

    @classmethod
    def get_module(cls, _backend):
        klass = super(InstanceModule, cls).get_module(_backend)
        return klass()



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
import sys
import tempfile
import time

import pytest


class NagiosReporter(object):

    def __init__(self):
        self.passed = 0
        self.failed = 0
        self.skipped = 0
        self.start_time = None
        self.total_time = None
        super(NagiosReporter, self).__init__()

    def pytest_sessionstart(self, session):
        self.start_time = time.time()

    def pytest_sessionfinish(self):
        self.total_time = time.time() - self.start_time

    def pytest_runtest_logreport(self, report):
        if report.passed:
            if report.when == "call":  # ignore setup/teardown
                self.passed += 1
        elif report.failed:
            self.failed += 1
        elif report.skipped:
            self.skipped += 1

    def report(self):
        if self.failed:
            status = "CRITICAL"
            ret = 2
        else:
            status = "OK"
            ret = 0

        sys.stdout.write((
            b"TESTINFRA %s - %d passed, %d failed, %d skipped in %.2f "
            b"seconds\n") % (
                status, self.passed, self.failed, self.skipped, self.total_time
            ))
        return ret


class RedirectStdStreams(object):
    # http://stackoverflow.com/questions/6796492/temporarily-redirect-stdout-stderr
    def __init__(self, stdout=None, stderr=None):
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr
        self._old_stdout = None
        self._old_stderr = None
        super(RedirectStdStreams, self).__init__()

    def __enter__(self):
        self._old_stdout, self._old_stderr = sys.stdout, sys.stderr
        self._old_stdout.flush()
        self._old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self._old_stdout
        sys.stderr = self._old_stderr


def main():
    if "--nagios" in sys.argv:
        out = tempfile.SpooledTemporaryFile()
        # Compat: In 2.7 SpooledTemporaryFile has no encoding param
        out.encoding = sys.stdout.encoding
        nagios_reporter = NagiosReporter()
        with RedirectStdStreams(stdout=out, stderr=out):
            pytest.main(plugins=[nagios_reporter])
        ret = nagios_reporter.report()
        out.seek(0)
        shutil.copyfileobj(out, sys.stdout)
        return ret
    else:
        return pytest.main()



# -*- coding: utf8 -*-
# Copyright © 2015 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from testinfra.modules.ansible import Ansible
from testinfra.modules.command import Command
from testinfra.modules.file import File
from testinfra.modules.group import Group
from testinfra.modules.interface import Interface
from testinfra.modules.package import Package
from testinfra.modules.process import Process
from testinfra.modules.puppet import Facter
from testinfra.modules.puppet import PuppetResource
from testinfra.modules.salt import Salt
from testinfra.modules.service import Service
from testinfra.modules.socket import Socket
from testinfra.modules.sysctl import Sysctl
from testinfra.modules.systeminfo import SystemInfo
from testinfra.modules.user import User


__all__ = [
    "Command", "File", "Package", "Group", "Interface",
    "Service", "SystemInfo", "User", "Salt", "PuppetResource",
    "Facter", "Sysctl", "Socket", "Ansible", "Process",
]



# -*- coding: utf8 -*-
# Copyright © 2016 Philippe Pepiot
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=import-error

from __future__ import unicode_literals
from __future__ import absolute_import

try:
    import ansible
except ImportError:
    _has_ansible = False
else:
    _has_ansible = True
    _ansible_major_version = int(ansible.__version__.split(".", 1)[0])
    if _ansible_major_version == 1:
        import ansible.inventory
        import ansible.runner
    elif _ansible_major_version == 2:
        import ansible.executor.task_queue_manager
        import ansible.inventory
        import ansible.parsing.dataloader
        import ansible.playbook.play
        import ansible.plugins.callback
        import ansible.vars


class AnsibleRunnerBase(object):

    def __init__(self, host_list=None):
        self.host_list = host_list
        super(AnsibleRunnerBase, self).__init__()

    def get_hosts(self, pattern=None):
        raise NotImplementedError

    def run(self, module_name, module_args, **kwargs):
        raise NotImplementedError


class AnsibleRunnerUnavailable(AnsibleRunnerBase):

    def get_hosts(self, pattern=None):
        raise RuntimeError(
            "You must install ansible package to use the ansible backend")

    def run(self, host, module_name, module_args, **kwargs):
        raise RuntimeError(
            "You must install ansible package to use the ansible backend")


class AnsibleRunnerV1(AnsibleRunnerBase):

    def __init__(self, host_list=None):
        super(AnsibleRunnerV1, self).__init__(host_list)
        self.inventory = ansible.inventory.Inventory(self.host_list)

    def get_hosts(self, pattern=None):
        return [
            e.name for e in
            self.inventory.get_hosts(pattern=pattern or "all")
        ]

    def run(self, host, module_name, module_args=None, **kwargs):
        kwargs = kwargs.copy()
        if self.host_list is not None:
            kwargs["host_list"] = self.host_list
        if module_args is not None:
            kwargs["module_args"] = module_args
        return ansible.runner.Runner(
            pattern=host,
            module_name=module_name,
            **kwargs).run()["contacted"][host]


class Options(object):

    def __init__(self, **kwargs):
        self.connection = "smart"
        for attr in (
            "module_path", "forks", "remote_user", "private_key_file",
            "ssh_common_args", "ssh_extra_args", "sftp_extra_args",
            "scp_extra_args", "become", "become_method", "become_user",
            "verbosity",
        ):
            setattr(self, attr, None)
        self.check = kwargs.get("check", False)
        super(Options, self).__init__()


if _has_ansible and _ansible_major_version == 2:
    class Callback(ansible.plugins.callback.CallbackBase):

        def __init__(self):
            self.unreachable = {}
            self.contacted = {}

        def runner_on_ok(self, host, result):
            self.contacted[host] = {
                "success": True,
                "result": result,
            }

        def runner_on_failed(self, host, result, ignore_errors=False):
            self.contacted[host] = {
                "success": False,
                "result": result,
            }

        def runner_on_unreachable(self, host, result):
            self.unreachable[host] = result


class AnsibleRunnerV2(AnsibleRunnerBase):

    def __init__(self, host_list=None):
        super(AnsibleRunnerV2, self).__init__(host_list)
        self.variable_manager = ansible.vars.VariableManager()
        self.loader = ansible.parsing.dataloader.DataLoader()
        self.inventory = ansible.inventory.Inventory(
            loader=self.loader,
            variable_manager=self.variable_manager,
            host_list=host_list,
        )
        self.variable_manager.set_inventory(self.inventory)

    def get_hosts(self, pattern=None):
        return [
            e.name for e in
            self.inventory.get_hosts(pattern=pattern or "all")
        ]

    def run(self, host, module_name, module_args=None, **kwargs):
        action = {"module": module_name}
        if module_args is not None:
            if module_name in ("command", "shell"):
                # Workaround https://github.com/ansible/ansible/issues/13862
                module_args = module_args.replace("=", "\\=")
            action["args"] = module_args
        play = ansible.playbook.play.Play().load({
            "hosts": host,
            "gather_facts": "no",
            "tasks": [{
                "action": action,
            }],
        }, variable_manager=self.variable_manager, loader=self.loader)
        tqm = None
        options = Options(**kwargs)
        callback = Callback()
        try:
            tqm = ansible.executor.task_queue_manager.TaskQueueManager(
                inventory=self.inventory,
                variable_manager=self.variable_manager,
                loader=self.loader,
                options=options,
                passwords=None,
                stdout_callback=callback,
            )
            tqm.run(play)
        finally:
            if tqm is not None:
                tqm.cleanup()

        if host in callback.unreachable:
            raise RuntimeError(
                "host unreachable: {}".format(callback.unreachable[host]))
        return callback.contacted[host]["result"]

if not _has_ansible:
    AnsibleRunner = AnsibleRunnerUnavailable
elif _ansible_major_version == 1:
    AnsibleRunner = AnsibleRunnerV1
elif _ansible_major_version == 2:
    AnsibleRunner = AnsibleRunnerV2
else:
    raise NotImplementedError(
        "Unhandled ansible version " + ansible.__version__
    )


def get_hosts(host_list=None, pattern=None):
    return AnsibleRunner(host_list).get_hosts(pattern)


def run(host, module_name, module_args=None, host_list=None, **kwargs):
    return AnsibleRunner(host_list).run(
        host, module_name, module_args, **kwargs)
